#include "dispatcher.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

/*
 * We currently support 10 threads at maximum
 */
#define MAX_THREADS 10

/*
 * The stack size of a single user-level thread (4 KiB = 1 page)
 */
#define STACK_SIZE 4096

typedef enum _ThreadState {
    STATE_UNUSED = 0, // This entry in the _threads array is unused.
    STATE_READY,      // The thread is ready to run
    STATE_RUNNING     // The thread is currently running. _currentThread will
                      // hold the thread's id.

    // No waiting state so far.
} ThreadState;

typedef struct _Thread {
    /*
     * Current state of the thread. It can be:
     */
    ThreadState state;

    /*
     * The same as the index in the threads array.
     */
    int threadId;

    /*
     * The stack pointer of this thread while it is not running.
     */
    void *currentSP;
} Thread;

/*
 * This holds the index/id of the currently running user-level thread.
 */
int _currentThread;

Thread _threads[MAX_THREADS];

/*
 * Called once from main() to initialize our user-level thread implementation
 */
void initThreads()
{
    memset(_threads, 0, sizeof(_threads));

    // We use the user-mode part of the current kernel-level thread as the
    // first user-level thread. We therefore, do not need to setup a stack,
    // because the thread already possesses a stack.
    _currentThread = 0;

    _threads[0].state    = STATE_RUNNING;
    _threads[0].threadId = 0;
    // Cannot set stack pointer, because thread 0 is still running (and the
    // stack pointer is thus changing constantly). The stack pointer can only
    // be set in yield().
}

/*
 * When a thread exits its main function, it usually has to be terminated.
 * That is, its data structures including its stack have to be freed and the
 * thread needs to be removed from the scheduling. To ease our world, we let
 * the thread fall into parkThread() when it exits its main function.
 * parkThread() then lets the thread indefinitely yield() so we do not need to
 * perform any clean up. We thus do not really terminate the thread.
 */
void _parkThread()
{
    while (1) yield();
}

/*
 * Update the _currentThread variable to whatever thread should be next. This
 * function implements the scheduling policy for the user-level threads.
 */
static void _scheduleNextThread()
{
    assert(_threads[_currentThread].state == STATE_RUNNING);

    _threads[_currentThread].state = STATE_READY;

    // Simple O(n) round-robin scheduling policy. Find the next ready thread
    // and set it to running.
    do {
        _currentThread++;
        _currentThread %= MAX_THREADS;
    } while (_threads[_currentThread].state != STATE_READY);
    _threads[_currentThread].state = STATE_RUNNING;
}

/*
 * Switches to the next thread.
 */
void yield()
{
    // Apply a scheduling policy to decide which user-level thread to run next.
    // This will update _currentThread to the next thread!
    int prevThread = _currentThread;
    _scheduleNextThread();

    printf("Switch from thread %d with sp near %p\n",
        prevThread, &prevThread);
    printf("Switch to thread %d with sp=%p\n",
        _currentThread, _threads[_currentThread].currentSP);

    assert(_threads[prevThread].state == STATE_READY);
    assert(_threads[_currentThread].state == STATE_RUNNING);
    assert(_threads[_currentThread].currentSP != NULL);

    // Do the context switch
    __asm__ __volatile__ (
        "pushq %%rbp\n\t"            // Push registers to old stack
        "pushq %%rbx\n\t"
        "pushq %%r12\n\t"
        "pushq %%r13\n\t"
        "pushq %%r14\n\t"
        "pushq %%r15\n\t"

        "movq  %%rsp, %[prevSp]\n\t" // Store current stack pointer
        "movq  %[newSp], %%rsp\n\t"  // Switch to the new stack

        "popq  %%r15\n\t"            // Pop registers from new stack
        "popq  %%r14\n\t"
        "popq  %%r13\n\t"
        "popq  %%r12\n\t"
        "popq  %%rbx\n\t"
        "popq  %%rbp"

        : [prevSp]"=m"(_threads[prevThread].currentSP)
        : [newSp]"m"(_threads[_currentThread].currentSP)
        : "cc", "memory");
}

/*
 * Prepares a new thread.
 * Returns the thread id or -1 on error.
 * This does not switch to the new thread, yet.
 */
int startThread(void (*func)(void))
{
    if (func == NULL) {
        return -1;
    }

    // Find a thread structure in our pre-allocated thread array, that we did
    // not use, yet.
    for (int i = 0; i < MAX_THREADS; ++i) {
        if (_threads[i].state == STATE_UNUSED) {

            // Allocate a stack for the user-level thread. For our purposes,
            // a simple, fixed-size memory region as stack is sufficient. We
            // can thus allocate stack space (i.e., memory space) from the
            // heap.
            void *stackBase = malloc(STACK_SIZE);
            if (stackBase == NULL) {
                return -1;
            }

            // Stacks grow from high addresses to low addresses. The stack thus
            // effectively starts at the end of the allocated memory area and
            // grows to the area's beginning.
            // We here assume the stack to be an array of void*'s, that is
            // pointers. A pointer has the right size (e.g., 32 or 64 bits) to
            // hold a full CPU register and can receive function pointers
            // without casting. stackTop is thus a pointer to pointers.
            void **stackTop = (void**)(stackBase + STACK_SIZE);

            // We need to write the values to the stack in the reverse order
            // they will be needed when yield() switches to this stack.

            // 'Push' the address that func() will pop from the stack and jump
            // to when it returns. We will use parkThread here.
            stackTop--;
            *stackTop = _parkThread;

            // The stack that is switched to on a yield().
            // The return of yield() will pop this address off the stack
            // and then jump to it. This way, the user function gets invoked.
            stackTop--;
            *stackTop = func;

            // Leave will pop this base pointer from the stack, when returning
            // from yield().
            stackTop--;
            *stackTop = stackTop;

            // Although we pop 6 registers after yield switches to this
            // stack, we only need to reserve space for 5 more, because we use
            // the position for the stack base pointer (ebp) twice (see
            // Understanding Solution PDF). We do not need to care about the
            // values because yield() switches to the thread for the first time
            // and there are thus no special values to restore.
            stackTop -= 5;

            // We initialized the stack and the thread is ready to run.
            // After setting STATE_READY, the thread is eligible to dispatching
            _threads[i].currentSP = stackTop;
            _threads[i].state     = STATE_READY;

            return i;
        }
    }

    return -1;
}

